@@ -2,6 +2,7 @@ package internal
 
 import (
 	log "github.com/sirupsen/logrus"
+	"github.com/walles/moor/v2/internal/search"
 	"github.com/walles/moor/v2/twin"
 )
 
@@ -23,14 +24,13 @@ func NewPagerModeFilter(p *Pager) *PagerModeFilter {
 	return m
 }
 
-func (m PagerModeFilter) drawFooter(_ string, _ string) {
+func (m PagerModeFilter) drawFooter(_ string, _ string, _ string) {
 	m.inputBox.draw(m.pager.screen, "Type to filter, 'ENTER' submits, 'ESC' cancels", "Filter: ")
 }
 
 func (m *PagerModeFilter) updateFilterPattern(text string) {
-	m.pager.filterPattern = toPattern(text)
-	m.pager.searchString = text
-	m.pager.searchPattern = toPattern(text)
+	m.pager.filter.For(text)
+	m.pager.search.For(text)
 }
 
 func (m *PagerModeFilter) onKey(key twin.KeyCode) {
@@ -44,9 +44,8 @@ func (m *PagerModeFilter) onKey(key twin.KeyCode) {
 
 	case twin.KeyEscape:
 		m.pager.mode = PagerModeViewing{pager: m.pager}
-		m.pager.filterPattern = nil
-		m.pager.searchString = ""
-		m.pager.searchPattern = nil
+		m.pager.filter = search.Search{}
+		m.pager.search.Clear()
 
 	case twin.KeyUp, twin.KeyDown, twin.KeyPgUp, twin.KeyPgDown:
 		viewing := PagerModeViewing{pager: m.pager}
