@@ -1,42 +1,27 @@
 package reader
 
 import (
-	"regexp"
-	"sync"
-
 	"github.com/walles/moor/v2/internal/linemetadata"
+	"github.com/walles/moor/v2/internal/search"
 	"github.com/walles/moor/v2/internal/textstyles"
 	"github.com/walles/moor/v2/twin"
 )
 
-// A Line represents a line of text that can / will be paged
-type Line struct {
-	raw   string
-	plain *string
-	lock  sync.Mutex
-}
-
-// NewLine creates a new Line from a (potentially ANSI / man page formatted) string
-func NewLine(raw string) Line {
-	return Line{
-		raw:   raw,
-		plain: nil,
-		lock:  sync.Mutex{},
-	}
-}
-
 // Returns a representation of the string split into styled tokens. Any regexp
 // matches are highlighted. A nil regexp means no highlighting.
+//
+// minRunesCount: at least this many runes will be included in the result. If 0,
+// do all runes. For BenchmarkRenderHugeLine() performance.
 func (line *Line) HighlightedTokens(
 	plainTextStyle twin.Style,
 	searchHitStyle twin.Style,
-	search *regexp.Regexp,
-	lineIndex *linemetadata.Index,
+	search search.Search,
+	lineIndex linemetadata.Index,
+	minRunesCount int,
 ) textstyles.StyledRunesWithTrailer {
-	plain := line.Plain(lineIndex)
-	matchRanges := getMatchRanges(&plain, search)
+	matchRanges := search.GetMatchRanges(line.Plain(lineIndex))
 
-	fromString := textstyles.StyledRunesFromString(plainTextStyle, line.raw, lineIndex)
+	fromString := textstyles.StyledRunesFromString(plainTextStyle, string(line.raw), &lineIndex, minRunesCount)
 	returnRunes := make([]textstyles.CellWithMetadata, 0, len(fromString.StyledRunes))
 	lastWasSearchHit := false
 	for _, token := range fromString.StyledRunes {
@@ -63,18 +48,6 @@ func (line *Line) HighlightedTokens(
 	}
 }
 
-// Plain returns a plain text representation of the initial string
-func (line *Line) Plain(lineIndex *linemetadata.Index) string {
-	line.lock.Lock()
-	defer line.lock.Unlock()
-
-	if line.plain == nil {
-		plain := textstyles.WithoutFormatting(line.raw, lineIndex)
-		line.plain = &plain
-	}
-	return *line.plain
-}
-
 func (line *Line) HasManPageFormatting() bool {
-	return textstyles.HasManPageFormatting(line.raw)
+	return textstyles.HasManPageFormatting(string(line.raw))
 }
