@@ -33,8 +33,9 @@ type renderedLine struct {
 
 type renderedScreen struct {
 	lines             []renderedLine
-	inputLines        []*reader.NumberedLine
+	inputLines        []reader.NumberedLine
 	numberPrefixWidth int // Including padding. 0 means no line numbers.
+	filenameText      string
 	statusText        string
 }
 
@@ -62,13 +63,13 @@ func (p *Pager) redraw(spinner string) {
 		// This happens when we're done
 		eofSpinner = "---"
 	}
-	spinnerLine := textstyles.StyledRunesFromString(statusbarStyle, eofSpinner, nil).StyledRunes
+	spinnerLine := textstyles.StyledRunesFromString(statusbarStyle, eofSpinner, nil, 0).StyledRunes
 	column := 0
 	for _, cell := range spinnerLine {
 		column += p.screen.SetCell(column, lastUpdatedScreenLineNumber+1, cell.ToStyledRune())
 	}
 
-	p.mode.drawFooter(renderedScreen.statusText, spinner)
+	p.mode.drawFooter(renderedScreen.filenameText, renderedScreen.statusText, spinner)
 
 	p.screen.Show()
 }
@@ -112,7 +113,7 @@ func (p *Pager) internalRenderLines(highlightSearchHitLines bool) renderedScreen
 	inputLines := p.Reader().GetLines(lineIndexToShow, p.visibleHeight())
 	if len(inputLines.Lines) == 0 {
 		// Empty input, empty output
-		return renderedScreen{statusText: inputLines.StatusText}
+		return renderedScreen{filenameText: inputLines.FilenameText, statusText: inputLines.StatusText}
 	}
 
 	lastVisibleLineNumber := inputLines.Lines[len(inputLines.Lines)-1].Number
@@ -195,6 +196,7 @@ func (p *Pager) internalRenderLines(highlightSearchHitLines bool) renderedScreen
 
 	return renderedScreen{
 		lines:             allLines,
+		filenameText:      inputLines.FilenameText,
 		statusText:        inputLines.StatusText,
 		inputLines:        inputLines.Lines,
 		numberPrefixWidth: numberPrefixLength,
@@ -208,13 +210,23 @@ func (p *Pager) internalRenderLines(highlightSearchHitLines bool) renderedScreen
 //
 // lineNumber and numberPrefixLength are required for knowing how much to
 // indent, and to (optionally) render the line number.
-func (p *Pager) renderLine(line *reader.NumberedLine, numberPrefixLength int, highlightSearchHitLines bool) []renderedLine {
-	highlighted := line.HighlightedTokens(plainTextStyle, searchHitStyle, p.searchPattern)
+func (p *Pager) renderLine(line reader.NumberedLine, numberPrefixLength int, highlightSearchHitLines bool) []renderedLine {
+	width, _ := p.screen.Size()
 	var wrapped []textstyles.StyledRunesWithTrailer
+	var highlighted textstyles.StyledRunesWithTrailer
 	if p.WrapLongLines {
-		width, _ := p.screen.Size()
+		highlighted = line.HighlightedTokens(plainTextStyle, searchHitStyle, p.search, 0)
+
 		wrapped = wrapLine(width-numberPrefixLength, highlighted.StyledRunes)
 	} else {
+		// Request only screen width tokens plus whatever is needed on the left
+		// due to horizontal scrolling. Also, get one extra to the right so we
+		// can know whether to show overflow markers.
+		//
+		// This is a huge performance gain when dealing with files with
+		// extremeny long lines: https://github.com/walles/moor/issues/358
+		highlighted = line.HighlightedTokens(plainTextStyle, searchHitStyle, p.search, width+p.leftColumnZeroBased+1)
+
 		// All on one line
 		wrapped = []textstyles.StyledRunesWithTrailer{{
 			StyledRunes:       highlighted.StyledRunes,
