@@ -1,7 +1,6 @@
 package internal
 
 import (
-	"regexp"
 	"strconv"
 	"strings"
 	"testing"
@@ -9,9 +8,12 @@ import (
 	"github.com/google/go-cmp/cmp"
 	"github.com/walles/moor/v2/internal/linemetadata"
 	"github.com/walles/moor/v2/internal/reader"
+	"github.com/walles/moor/v2/internal/search"
 	"github.com/walles/moor/v2/internal/textstyles"
 	"github.com/walles/moor/v2/twin"
 	"gotest.tools/v3/assert"
+
+	log "github.com/sirupsen/logrus"
 )
 
 // NOTE: You can find related tests in pager_test.go.
@@ -27,7 +29,12 @@ func renderedToString(row []textstyles.CellWithMetadata) string {
 }
 
 func testHorizontalCropping(t *testing.T, contents string, firstVisibleColumn int, lastVisibleColumn int, expected string) {
-	pager := NewPager(nil)
+	log.SetLevel(log.WarnLevel) // Stop info logs from polluting benchmark output
+
+	reader := reader.NewFromTextForTesting("testHorizontalCropping", contents)
+	assert.NilError(t, reader.Wait())
+
+	pager := NewPager(reader)
 	pager.ShowLineNumbers = false
 	pager.showLineNumbers = false
 
@@ -35,11 +42,10 @@ func testHorizontalCropping(t *testing.T, contents string, firstVisibleColumn in
 	pager.leftColumnZeroBased = firstVisibleColumn
 	pager.scrollPosition = newScrollPosition("testHorizontalCropping")
 
-	lineContents := reader.NewLine(contents)
-	numberedLine := reader.NumberedLine{
-		Line: &lineContents,
-	}
-	screenLine := pager.renderLine(&numberedLine, pager.getLineNumberPrefixLength(numberedLine.Number), true)
+	numberedLine := reader.GetLine(linemetadata.IndexFromZeroBased(0))
+	assert.Assert(t, numberedLine != nil)
+
+	screenLine := pager.renderLine(*numberedLine, pager.getLineNumberPrefixLength(numberedLine.Number), true)
 	assert.Equal(t, renderedToString(screenLine[0].cells), expected)
 }
 
@@ -98,27 +104,25 @@ func TestEmpty(t *testing.T) {
 	}
 	pager.filteringReader = FilteringReader{
 		BackingReader: pager.readers[pager.currentReader],
-		FilterPattern: &pager.filterPattern,
+		Filter:        &pager.filter,
 	}
 
 	rendered := pager.renderLines()
 	assert.Equal(t, len(rendered.lines), 0)
-	assert.Equal(t, "test: <empty>", rendered.statusText)
+	assert.Equal(t, "test", rendered.filenameText)
+	assert.Equal(t, ": <empty>", rendered.statusText)
 	assert.Assert(t, pager.lineIndex() == nil)
 }
 
 // Repro case for a search bug discovered in v1.9.8.
 func TestSearchHighlight(t *testing.T) {
-	line := reader.NewLine("x\"\"x")
+	numberedLine := reader.NewFromTextForTesting("TestSearchHighlight", "x\"\"x").GetLine(linemetadata.Index{})
 	pager := Pager{
-		screen:        twin.NewFakeScreen(100, 10),
-		searchPattern: regexp.MustCompile("\""),
+		screen: twin.NewFakeScreen(100, 10),
+		search: search.For("\""),
 	}
 
-	numberedLine := reader.NumberedLine{
-		Line: &line,
-	}
-	rendered := pager.renderLine(&numberedLine, pager.getLineNumberPrefixLength(numberedLine.Number), true)
+	rendered := pager.renderLine(*numberedLine, pager.getLineNumberPrefixLength(numberedLine.Number), true)
 	assert.DeepEqual(t, []renderedLine{
 		{
 			inputLineIndex:    linemetadata.Index{},
@@ -154,13 +158,14 @@ func TestOverflowDown(t *testing.T) {
 	}
 	pager.filteringReader = FilteringReader{
 		BackingReader: pager.readers[pager.currentReader],
-		FilterPattern: &pager.filterPattern,
+		Filter:        &pager.filter,
 	}
 
 	rendered := pager.renderLines()
 	assert.Equal(t, len(rendered.lines), 1)
 	assert.Equal(t, "hej", renderedToString(rendered.lines[0].cells))
-	assert.Equal(t, "test: 1 line  100%", rendered.statusText)
+	assert.Equal(t, "test", rendered.filenameText)
+	assert.Equal(t, ": 1 line  100%", rendered.statusText)
 	assert.Assert(t, pager.lineIndex().IsZero())
 	assert.Equal(t, pager.deltaScreenLines(), 0)
 }
@@ -179,13 +184,14 @@ func TestOverflowUp(t *testing.T) {
 	}
 	pager.filteringReader = FilteringReader{
 		BackingReader: pager.readers[pager.currentReader],
-		FilterPattern: &pager.filterPattern,
+		Filter:        &pager.filter,
 	}
 
 	rendered := pager.renderLines()
 	assert.Equal(t, len(rendered.lines), 1)
 	assert.Equal(t, "hej", renderedToString(rendered.lines[0].cells))
-	assert.Equal(t, "test: 1 line  100%", rendered.statusText)
+	assert.Equal(t, "test", rendered.filenameText)
+	assert.Equal(t, ": 1 line  100%", rendered.statusText)
 	assert.Assert(t, pager.lineIndex().IsZero())
 	assert.Equal(t, pager.deltaScreenLines(), 0)
 }
@@ -249,7 +255,7 @@ func TestOneLineTerminal(t *testing.T) {
 	}
 	pager.filteringReader = FilteringReader{
 		BackingReader: pager.readers[pager.currentReader],
-		FilterPattern: &pager.filterPattern,
+		Filter:        &pager.filter,
 	}
 
 	rendered := pager.renderLines()
@@ -276,14 +282,14 @@ func TestShortenedInput(t *testing.T) {
 
 	pager.filteringReader = FilteringReader{
 		BackingReader: pager.readers[pager.currentReader],
-		FilterPattern: &pager.filterPattern,
+		Filter:        &pager.filter,
 	}
 
 	pager.scrollToEnd()
 	assert.Equal(t, pager.lineIndex().Index(), 991, "This should have been the effect of calling scrollToEnd()")
 
 	pager.mode = NewPagerModeFilter(&pager)
-	pager.filterPattern = regexp.MustCompile("first") // Match only the first line
+	pager.filter = search.For("first") // Match only the first line
 
 	rendered := pager.renderLines()
 	assert.Equal(t, len(rendered.lines), 1, "Should have rendered one line")
@@ -314,14 +320,14 @@ func TestShortenedInputManyLines(t *testing.T) {
 
 	pager.filteringReader = FilteringReader{
 		BackingReader: pager.readers[pager.currentReader],
-		FilterPattern: &pager.filterPattern,
+		Filter:        &pager.filter,
 	}
 
 	pager.scrollToEnd()
 	assert.Equal(t, pager.lineIndex().Index(), 991, "Should be at the last line before filtering")
 
 	pager.mode = NewPagerModeFilter(&pager)
-	pager.filterPattern = regexp.MustCompile(`^match`)
+	pager.filter = search.For(`^match`)
 
 	rendered := pager.renderLines()
 	assert.Equal(t, len(rendered.lines), 9, "Should have rendered 9 lines (10 minus one status bar)")
@@ -391,9 +397,9 @@ func testRenderLinesWithSearchHits(t *testing.T, input string, expectedBackgroun
 	}
 	pager.filteringReader = FilteringReader{
 		BackingReader: pager.readers[pager.currentReader],
-		FilterPattern: &pager.filterPattern,
+		Filter:        &pager.filter,
 	}
-	pager.searchPattern = regexp.MustCompile("xxx")
+	pager.search = search.For("xxx")
 	pager.ShowStatusBar = false
 	pager.mode = PagerModeViewing{&pager}
 	pager.showLineNumbers = false
@@ -424,3 +430,28 @@ func BenchmarkRenderLines(b *testing.B) {
 		pager.renderLines()
 	}
 }
+
+// Inspired by https://github.com/walles/moor/issues/358
+func BenchmarkRenderHugeLine(b *testing.B) {
+	const megabytes = 5
+	builder := strings.Builder{}
+	for builder.Len() < megabytes*1024*1024 {
+		builder.WriteString("Romani ite domum. ")
+	}
+	b.SetBytes(int64(builder.Len()))
+
+	input := reader.NewFromTextForTesting(
+		"BenchmarkRenderHugeLine()",
+		builder.String())
+	pager := NewPager(input)
+	pager.screen = twin.NewFakeScreen(80, 25)
+
+	assert.NilError(b, input.Wait())
+
+	pager.renderLines() // Warm up
+
+	b.ResetTimer()
+	for i := 0; i < b.N; i++ {
+		pager.renderLines()
+	}
+}
