@@ -19,6 +19,8 @@ var lineNumbersStyle = twin.StyleDefault.WithAttr(twin.AttrDim)
 // Status bar and EOF marker style
 var statusbarStyle = twin.StyleDefault.WithAttr(twin.AttrReverse)
 
+var statusbarFileStyle = twin.StyleDefault.WithAttr(twin.AttrReverse).WithAttr(twin.AttrUnderline)
+
 var plainTextStyle = twin.StyleDefault
 
 var searchHitStyle = twin.StyleDefault.WithAttr(twin.AttrReverse)
@@ -62,7 +64,7 @@ func twinStyleFromChroma(terminalBackground *twin.Color, chromaStyle *chroma.Sty
 	}
 
 	formatted := stringBuilder.String()
-	cells := textstyles.StyledRunesFromString(twin.StyleDefault, formatted, nil).StyledRunes
+	cells := textstyles.StyledRunesFromString(twin.StyleDefault, formatted, nil, 0).StyledRunes
 	if len(cells) != 1 {
 		log.Warnf("Chroma formatter didn't return exactly one cell: %#v", cells)
 		return nil
@@ -138,7 +140,7 @@ func getOppositeColor(base twin.Color) twin.Color {
 	}
 }
 
-func styleUI(terminalBackground *twin.Color, chromaStyle *chroma.Style, chromaFormatter *chroma.Formatter, statusbarOption StatusBarOption, withTerminalFg bool) {
+func styleUI(terminalBackground *twin.Color, chromaStyle *chroma.Style, chromaFormatter *chroma.Formatter, statusbarOption StatusBarOption, withTerminalFg bool, configureSearchHitLineBackground bool) {
 	// Set defaults
 	plainTextStyle = twin.StyleDefault
 	textstyles.ManPageHeading = twin.StyleDefault.WithAttr(twin.AttrBold)
@@ -192,12 +194,14 @@ func styleUI(terminalBackground *twin.Color, chromaStyle *chroma.Style, chromaFo
 		panic(fmt.Sprint("Unrecognized status bar style: ", statusbarOption))
 	}
 
-	configureHighlighting(terminalBackground)
+	statusbarFileStyle = statusbarStyle.WithAttr(twin.AttrUnderline)
+
+	configureHighlighting(terminalBackground, configureSearchHitLineBackground)
 }
 
 // Expects to be called from the end of styleUI(), since at that
 // point we should have all data we need to set up highlighting.
-func configureHighlighting(terminalBackground *twin.Color) {
+func configureHighlighting(terminalBackground *twin.Color, configureSearchHitLineBackground bool) {
 	if standoutStyle != nil {
 		searchHitStyle = *standoutStyle
 		log.Trace("Search hit style set from standout style: ", searchHitStyle)
@@ -205,7 +209,16 @@ func configureHighlighting(terminalBackground *twin.Color) {
 		log.Trace("Search hit style set to default: ", searchHitStyle)
 	}
 
-	// Figure out a line background that lies between plainTextStyle and searchHit
+	//
+	// Everything below this point relates to figuring out which background
+	// color we should use for lines with search hits.
+	//
+
+	if !configureSearchHitLineBackground {
+		log.Trace("Not configuring search hit line background color")
+		return
+	}
+
 	var plainBg twin.Color
 	if terminalBackground != nil {
 		plainBg = *terminalBackground
@@ -249,7 +262,7 @@ func configureHighlighting(terminalBackground *twin.Color) {
 
 func TermcapToStyle(termcap string) (twin.Style, error) {
 	// Add a character to be sure we have one to take the format from
-	cells := textstyles.StyledRunesFromString(twin.StyleDefault, termcap+"x", nil).StyledRunes
+	cells := textstyles.StyledRunesFromString(twin.StyleDefault, termcap+"x", nil, 0).StyledRunes
 	if len(cells) != 1 {
 		return twin.StyleDefault, fmt.Errorf("Expected styling only and no text")
 	}
